@@ -4,6 +4,7 @@
 
 from __future__ import print_function, absolute_import, unicode_literals
 
+import bisect
 import contextlib
 import copy
 import errno
@@ -698,37 +699,50 @@ class ModulesManager(object):
                 pass
 
             search_path = self._get_platform_file_path(module_path)
+
             for _, value in list(sys.modules.items()):
                 try:
                     path = self._get_platform_file_path(value.__file__)
                 except AttributeError:
                     path = None
 
-                if path and search_path == path:
-                    module_id = self._next_id
-                    self._next_id += 1
+                if not path:
+                    continue
 
-                    module = {
-                        'id': module_id,
-                        'package': value.__package__ if hasattr(value, '__package__') else None,
-                        'path': module_path,
-                    }
+                try:
+                    # This tries to open the files to obtain handles, which can be restricted
+                    # by file permissions, but ensures that long/short path mismatch, symlinks
+                    # etc are all accounted for. Fall back to comparing names in case of failure.
+                    if not os.path.samefile(path, search_path):
+                        continue
+                except Exception:
+                    if path != search_path:
+                        continue
 
-                    try:
-                        module['name'] = value.__qualname__
-                    except AttributeError:
-                        module['name'] = value.__name__
+                module_id = self._next_id
+                self._next_id += 1
 
-                    try:
-                        module['version'] = value.__version__
-                    except AttributeError:
-                        pass
+                module = {
+                    'id': module_id,
+                    'package': value.__package__ if hasattr(value, '__package__') else None,
+                    'path': module_path,
+                }
 
-                    self.path_to_module_id[module_path] = module_id
-                    self.module_id_to_details[module_id] = module
+                try:
+                    module['name'] = value.__qualname__
+                except AttributeError:
+                    module['name'] = value.__name__
 
-                    self.proc.send_event('module', reason='new', module=module)
-                    return module
+                try:
+                    module['version'] = value.__version__
+                except AttributeError:
+                    pass
+
+                self.path_to_module_id[module_path] = module_id
+                self.module_id_to_details[module_id] = module
+
+                self.proc.send_event('module', reason='new', module=module)
+                return module
 
         return None
 
@@ -806,6 +820,7 @@ def bool_parser(str):
 DEBUG_OPTIONS_PARSER = {
     'WAIT_ON_ABNORMAL_EXIT': bool_parser,
     'WAIT_ON_NORMAL_EXIT': bool_parser,
+    'BREAK_SYSTEMEXIT_ZERO': bool_parser,
     'REDIRECT_OUTPUT': bool_parser,
     'VERSION': unquote,
     'INTERPRETER_OPTIONS': unquote,
@@ -824,6 +839,7 @@ DEBUG_OPTIONS_BY_FLAG = {
     'RedirectOutput': 'REDIRECT_OUTPUT=True',
     'WaitOnNormalExit': 'WAIT_ON_NORMAL_EXIT=True',
     'WaitOnAbnormalExit': 'WAIT_ON_ABNORMAL_EXIT=True',
+    'BreakOnSystemExitZero': 'BREAK_SYSTEMEXIT_ZERO=True',
     'Django': 'DJANGO_DEBUG=True',
     'Flask': 'FLASK_DEBUG=True',
     'Jinja': 'FLASK_DEBUG=True',
@@ -877,6 +893,7 @@ def _parse_debug_options(opts):
     """Debug options are semicolon separated key=value pairs
         WAIT_ON_ABNORMAL_EXIT=True|False
         WAIT_ON_NORMAL_EXIT=True|False
+        BREAK_SYSTEMEXIT_ZERO=True|False
         REDIRECT_OUTPUT=True|False
         VERSION=string
         INTERPRETER_OPTIONS=string
@@ -1200,7 +1217,7 @@ class VSCLifecycleMsgProcessor(VSCodeMessageProcessorBase):
         multiproc.root_start_request = request
         self.start_reason = 'attach'
         self._set_debug_options(args)
-        self._handle_attach(request, args)
+        self._handle_launch_or_attach(request, args)
         self.send_response(request)
 
     def on_launch(self, request, args):
@@ -1208,7 +1225,7 @@ class VSCLifecycleMsgProcessor(VSCodeMessageProcessorBase):
         self.start_reason = 'launch'
         self._set_debug_options(args)
         self._notify_launch()
-        self._handle_launch(request, args)
+        self._handle_launch_or_attach(request, args)
         self.send_response(request)
 
     def on_disconnect(self, request, args):
@@ -1289,10 +1306,7 @@ class VSCLifecycleMsgProcessor(VSCodeMessageProcessorBase):
     def _handle_configurationDone(self, request, args):
         pass
 
-    def _handle_attach(self, request, args):
-        pass
-
-    def _handle_launch(self, request, args):
+    def _handle_launch_or_attach(self, request, args):
         pass
 
     def _handle_detach(self):
@@ -1343,6 +1357,8 @@ class VSCodeMessageProcessor(VSCLifecycleMsgProcessor):
         # adapter state
         self.path_casing = PathUnNormcase()
         self._detached = False
+        self._path_mappings_received = False
+        self._path_mappings_applied = False
 
     def _start_event_loop(self):
         self.loop = futures.EventLoop()
@@ -1596,6 +1612,8 @@ class VSCodeMessageProcessor(VSCLifecycleMsgProcessor):
         if len(self._path_mappings) > 0:
             pydevd_file_utils.setup_client_server_paths(self._path_mappings)
 
+        self._path_mappings_applied = True
+
     def _send_cmd_version_command(self):
         cmd = pydevd_comm.CMD_VERSION
         default_os_type = 'WINDOWS' if platform.system() == 'Windows' else 'UNIX'
@@ -1605,18 +1623,9 @@ class VSCodeMessageProcessor(VSCLifecycleMsgProcessor):
         return self.pydevd_request(cmd, msg)
 
     @async_handler
-    def _handle_attach(self, request, args):
-        self.pydevd_request(pydevd_comm.CMD_SET_PROTOCOL, 'json')
-        yield self._send_cmd_version_command()
+    def _handle_launch_or_attach(self, request, args):
+        self._path_mappings_received = True
 
-        pydevd_request = copy.deepcopy(request)
-        del pydevd_request['seq']  # A new seq should be created for pydevd.
-        yield self.pydevd_request(-1, pydevd_request, is_json=True)
-
-        self._initialize_path_maps(args)
-
-    @async_handler
-    def _handle_launch(self, request, args):
         self.pydevd_request(pydevd_comm.CMD_SET_PROTOCOL, 'json')
         yield self._send_cmd_version_command()
 
@@ -2263,8 +2272,51 @@ class VSCodeMessageProcessor(VSCLifecycleMsgProcessor):
 
     @async_handler
     def on_setBreakpoints(self, request, args):
+        if not self._path_mappings_received:
+            self.send_error_response(request, "'setBreakpoints' request must be issued after 'launch' or 'attach' request.")
+            return
+
+        # There might be a concurrent 'launch' or 'attach' request in flight that hasn't
+        # gotten to processing path mappings yet. If so, spin until it finishes that.
+        while not self._path_mappings_applied:
+            yield self.sleep()
+
         pydevd_request = copy.deepcopy(request)
         del pydevd_request['seq']  # A new seq should be created for pydevd.
+
+        # Validate breakpoints and adjust their positions.
+        args = pydevd_request['arguments']
+        if 'breakpoints' in args and 'path' in args['source']:
+            path = args['source']['path']
+            if sys.version_info < (3,) and not isinstance(path, bytes):
+                path = path.encode(sys.getfilesystemencoding())
+            path = path_to_unicode(pydevd_file_utils.norm_file_to_server(path))
+
+            try:
+                with open(path) as f:
+                    src = f.read()
+            except Exception:
+                pass
+            else:
+                try:
+                    code = compile(src, path, 'exec', 0, dont_inherit=True)
+                except Exception:
+                    pass
+                else:
+                    try:
+                        lines = sorted(_util.get_code_lines(code))
+                    except Exception:
+                        pass
+                    else:
+                        if lines:
+                            for bp in args['breakpoints']:
+                                line = bp['line']
+                                if line not in lines:
+                                    # Adjust to the first preceding valid line.
+                                    idx = bisect.bisect_left(lines, line)
+                                    if idx > 0:
+                                        bp['line'] = lines[idx - 1]
+
         _, _, resp_args = yield self.pydevd_request(
             pydevd_comm.CMD_SET_BREAK,
             pydevd_request,
@@ -2556,8 +2608,23 @@ class VSCodeMessageProcessor(VSCLifecycleMsgProcessor):
             exc_name, exc_desc, _, _ = \
                 self._parse_exception_details(resp_args, include_stack=False)
 
-        extra['allThreadsStopped'] = True
+        if not self.debug_options.get('BREAK_SYSTEMEXIT_ZERO', False):
+            # SystemExit is qualified on Python 2, and unqualified on Python 3
+            sysexit_exc_name = 'exceptions.SystemExit' if sys.version_info < (3,) else 'SystemExit'
+            if exc_name == sysexit_exc_name:
+                try:
+                    exit_code = int(exc_desc)
+                except ValueError:
+                    # It is legal to invoke exit() with a non-integer argument, and SystemExit will
+                    # pass that through. It's considered an error exit, same as non-zero integer.
+                    ignore = False
+                else:
+                    ignore = (exit_code == 0)
+                if ignore:
+                    self._resume_all_threads()
+                    return
 
+        extra['allThreadsStopped'] = True
         self.send_event(
             'stopped',
             reason=reason,
